#include "stdafx.h"
#include <iostream>
#include  <iomanip>
#include <Windows.h>
using namespace std;

#define PAUSE_AND_RETURN system("pause"); return 0
#define _KB *1024
#define _MB *1024*1024
#define _GB *1024*1024*1024
typedef DWORD u32;

BOOL WriteFileData(HANDLE hFile, u32 startOffset, byte* dataBuffer,  u32 dataSize){
	u32 currentOffset = startOffset;
	u32 endOffset = startOffset+dataSize;
	u32 mBufferSize = (dataSize < 64 _KB) ? dataSize : 64 _KB;
	u32 byteswritten = 0;
	SetFilePointer(hFile,startOffset, NULL, FILE_BEGIN);
	while(1){
		if(!WriteFile(hFile,dataBuffer + (currentOffset - startOffset), mBufferSize,&byteswritten, NULL)){
			return FALSE;
		}
		currentOffset+=mBufferSize;
		if(currentOffset >= endOffset)
			break;
	}
	return TRUE;
}

BOOL ReadFileData(HANDLE hFile, u32 startOffset, byte* dataBuffer,  u32 dataSize){
	u32 currentOffset = startOffset;
	u32 endOffset = startOffset+dataSize;
	u32 mBufferSize = (dataSize < 64 _KB) ? dataSize : 64 _KB;
	u32 bytesread = 0;
	SetFilePointer(hFile,startOffset, NULL, FILE_BEGIN);
	while(1){
		if(!ReadFile(hFile,dataBuffer + (currentOffset - startOffset), mBufferSize, &bytesread, NULL)){
			return FALSE;
		}
		currentOffset+=mBufferSize;
		if(currentOffset >= endOffset)
			break;
	}
	return TRUE;
}

u32 GetSaveDataAndSize(HANDLE hFile, u32 saveOffset, byte* saveDataBuffer, u32 bufferSize){
	if(ReadFileData(hFile,saveOffset, saveDataBuffer,bufferSize)){
		for (u32 i = bufferSize - 1; i > 0; i--){
			if(saveDataBuffer[i] != 0xFF && saveDataBuffer[i] != 0x00){
				u32 correctedSize = 1 _MB;
				double n = 20;
				while(correctedSize < i){
					n++;
					correctedSize = (u32)pow(2,n);
				}
				return correctedSize;
			}
		}
	}
	return 0;
}

int _tmain(int argc, _TCHAR* argv[])
{
	//Check parameter, file format, etc.
	if(argc < 2){
		cout << ">> Error: Incorrect parameters\nUsage: \nCard2RomTrimTool [filename] [Optional: savedataspace in MB]\n\n";
		PAUSE_AND_RETURN;
	}
	if(_tcsicmp((TCHAR*)(argv[1]+_tcslen(argv[1]) - 4), L".3ds") != 0){
		cout << ">> Error: Only .3ds file is supported!\n\n";
		PAUSE_AND_RETURN;
	}
	HANDLE hFile = CreateFile(argv[1], GENERIC_READ|GENERIC_WRITE,NULL,NULL,OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if(hFile == INVALID_HANDLE_VALUE){
		wcout << L">> Error: Cannot open " << argv[1] << L"\n\n";
		PAUSE_AND_RETURN;
	}
	//Check whether this is a card2 rom file.
	u32 oldSaveOffset = 0xFFFFFFFF;
	ReadFileData(hFile, 0x200, (byte*)&oldSaveOffset, sizeof u32);
	if(oldSaveOffset == 0xFFFFFFFF){
		cout << ">> Error: This is not a Card2 game!\n\n";
		CloseHandle(hFile); 
		PAUSE_AND_RETURN;
	}
	else{
		oldSaveOffset*=0x200;
	}
	//Get the actual rom size.
	u32 romActualSize = 0;
	if(!ReadFileData(hFile, 0x300, (byte*)&romActualSize, sizeof u32)){
		cout << ">> Error: Cannot read partition table, undable to determine actual rom size.\n\n";
		CloseHandle(hFile); 
		PAUSE_AND_RETURN;
	}
	//Scan for existing save data
	cout << "Scanning for existing save data...\n\n";
	u32 writableAreaSize = GetFileSize(hFile,0) - oldSaveOffset;

	u32 saveDataSize = (writableAreaSize > 16 _MB) ? 16 _MB : writableAreaSize;
	if(argc == 3){
		u32 sizeInMB = _tstoi(argv[2]);
		saveDataSize = sizeInMB _MB;
	}

	byte* tempData = new byte[saveDataSize];
	u32 existingSaveDataSize = GetSaveDataAndSize(hFile, oldSaveOffset, tempData, saveDataSize);
	if(existingSaveDataSize == 0){
		cout << ">> Error: No existing save data found.\n";
		cout << ">> You need existing save data to avoid problems during save initialization.\n>> Please load the rom at least once before trimming.\n\n";
		CloseHandle(hFile); 
		delete[] tempData;
		PAUSE_AND_RETURN;
	}
	cout << "Found existing save data at 0x" << hex << uppercase << setfill('0') << setw(8) << oldSaveOffset << dec << " Size: " << (double)existingSaveDataSize/1024/1024 << "MB\n";
	//Transfer existing save data to new save offset.
	u32 newSaveOffset = romActualSize + 0x200;
	if(newSaveOffset >= oldSaveOffset){
		newSaveOffset = oldSaveOffset;
		cout << "\n";
	}
	else
	{
		cout << "Transfering save data to new location...\n\n";
		WriteFileData(hFile, newSaveOffset, tempData, existingSaveDataSize);
	}
	delete[] tempData;
	//Change original save offset to new save offset.
	u32 saveOffsetMediaUnit = newSaveOffset/0x200;
	WriteFileData(hFile, 0x200, (byte*)&saveOffsetMediaUnit, sizeof u32);
	//Setup save data space, 16 MB by default, unless user specify another value.

	//Tim the file.
	if(newSaveOffset+saveDataSize >= GetFileSize(hFile,NULL)){
		cout << ">> Warning: This rom may have already been trimmed.\n>> New file size is equal or larger than current file size. \n";
		cout << ">> Trimming will be skipped.\n\n";
		saveDataSize = GetFileSize(hFile, NULL) - newSaveOffset;
	} else {
		SetFilePointer(hFile,newSaveOffset+saveDataSize,0,FILE_BEGIN);
		SetEndOfFile(hFile);
	}
	//Done.
	cout << "Done!\n";
	cout << "Save offset is set to: 0x" << hex << uppercase << setfill('0') << setw(8) << newSaveOffset << "\n";
	cout << "Save data space: " << dec << (double)saveDataSize/1024/1024 << "MB (" << (double)existingSaveDataSize/1024/1024 << "MB used)\n";
	cout << "Rom file size: " <<fixed << setprecision(2) << (double)(GetFileSize(hFile,NULL)) /1024/1024 << "MB\n\n";
	CloseHandle(hFile);
	PAUSE_AND_RETURN;
}